"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from functools import partial
from typing import Callable, Optional, Type, Union

import numpy as np
import torch
import torch.nn as nn


def get_device(device: Optional[Union[str, torch.device]] = None) -> torch.device:
    """
    Instantiates a PyTorch device object.

    Parameters
    ----------
    device
        Either `None`, a str ('gpu' or 'cpu') indicating the device to choose, or an
        already instantiated device object. If `None`, the GPU is selected if it is
        detected, otherwise the CPU is used as a fallback.

    Returns
    -------
    The instantiated device object.
    """
    if isinstance(device, torch.device):  # Already a torch device
        return device
    else:  # Instantiate device
        if device is None or device.lower() in ["gpu", "cuda"]:
            torch_device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        else:
            torch_device = torch.device("cpu")
    return torch_device


def mmd2_from_kernel_matrix(
    kernel_mat: torch.Tensor, m: int, permute: bool = False, zero_diag: bool = True
) -> torch.Tensor:
    """
    Compute maximum mean discrepancy (MMD^2) between 2 samples x and y from the
    full kernel matrix between the samples.

    Parameters
    ----------
    kernel_mat
        Kernel matrix between samples x and y.
    m
        Number of instances in y.
    permute
        Whether to permute the row indices. Used for permutation tests.
    zero_diag
        Whether to zero out the diagonal of the kernel matrix.

    Returns
    -------
    MMD^2 between the samples from the kernel matrix.
    """
    n = kernel_mat.shape[0] - m
    if zero_diag:
        kernel_mat = kernel_mat - torch.diag(kernel_mat.diag())
    if permute:
        idx = torch.randperm(kernel_mat.shape[0])
        kernel_mat = kernel_mat[idx][:, idx]
    k_xx, k_yy, k_xy = kernel_mat[:-m, :-m], kernel_mat[-m:, -m:], kernel_mat[-m:, :-m]
    c_xx, c_yy = 1 / (n * (n - 1)), 1 / (m * (m - 1))
    mmd2 = c_xx * k_xx.sum() + c_yy * k_yy.sum() - 2.0 * k_xy.mean()
    return mmd2


def predict_batch(
    x: Union[np.ndarray, torch.Tensor],
    model: Union[Callable, nn.Module, nn.Sequential],
    device: Optional[torch.device] = None,
    batch_size: int = int(1e10),
    preprocess_fn: Optional[Callable] = None,
    dtype: Union[Type[np.generic], torch.dtype] = np.float32,
) -> Union[np.ndarray, torch.Tensor, tuple]:
    """
    Make batch predictions on a model.

    Parameters
    ----------
    x
        Batch of instances.
    model
        PyTorch model.
    device
        Device type used. The default None tries to use the GPU and falls back on CPU.
        Can be specified by passing either torch.device('cuda') or torch.device('cpu').
    batch_size
        Batch size used during prediction.
    preprocess_fn
        Optional preprocessing function for each batch.
    dtype
        Model output type, e.g. np.float32 or torch.float32.

    Returns
    -------
    Numpy array, torch tensor or tuples of those with model outputs.
    """
    device = get_device(device)
    if isinstance(x, np.ndarray):
        x = torch.from_numpy(x)
    n = len(x)
    n_minibatch = int(np.ceil(n / batch_size))
    return_np = not isinstance(dtype, torch.dtype)
    preds = []
    with torch.no_grad():
        for i in range(n_minibatch):
            istart, istop = i * batch_size, min((i + 1) * batch_size, n)
            x_batch = x[istart:istop]
            if isinstance(preprocess_fn, Callable):
                x_batch = preprocess_fn(x_batch)
            preds_tmp = model(x_batch.to(device))
            if isinstance(preds_tmp, (list, tuple)):
                if len(preds) == 0:  # init tuple with lists to store predictions
                    preds = tuple([] for _ in range(len(preds_tmp)))
                for j, p in enumerate(preds_tmp):
                    if isinstance(p, torch.Tensor):
                        p = p.cpu()
                    preds[j].append(p if not return_np or isinstance(p, np.ndarray) else p.numpy())
            elif isinstance(preds_tmp, (np.ndarray, torch.Tensor)):
                if isinstance(preds_tmp, torch.Tensor):
                    preds_tmp = preds_tmp.cpu()
                if isinstance(preds, tuple):
                    preds = list(preds)
                preds.append(
                    preds_tmp
                    if not return_np or isinstance(preds_tmp, np.ndarray)  # type: ignore
                    else preds_tmp.numpy()
                )
            else:
                raise TypeError(
                    f"Model output type {type(preds_tmp)} not supported. The model \
                    output type needs to be one of list, tuple, np.ndarray or \
                    torch.Tensor."
                )
    concat = partial(np.concatenate, axis=0) if return_np else partial(torch.cat, dim=0)
    out: Union[tuple, np.ndarray, torch.Tensor] = (
        tuple(concat(p) for p in preds) if isinstance(preds, tuple) else concat(preds)  # type: ignore
    )
    return out


def preprocess_drift(
    x: np.ndarray,
    model: nn.Module,
    device: Optional[torch.device] = None,
    preprocess_batch_fn: Optional[Callable] = None,
    batch_size: int = int(1e10),
    dtype: Union[Type[np.generic], torch.dtype] = np.float32,
) -> Union[np.ndarray, torch.Tensor, tuple]:
    """
    Prediction function used for preprocessing step of drift detector.

    Parameters
    ----------
    x
        Batch of instances.
    model
        Model used for preprocessing.
    device
        Device type used. The default None tries to use the GPU and falls back on CPU.
        Can be specified by passing either torch.device('cuda') or torch.device('cpu').
    preprocess_batch_fn
        Optional batch preprocessing function. For example to convert a list of objects
        to a batch which can be processed by the PyTorch model.
    batch_size
        Batch size used during prediction.
    dtype
        Model output type, e.g. np.float32 or torch.float32.

    Returns
    -------
    Numpy array or torch tensor with predictions.
    """
    return predict_batch(
        x,
        model,
        device=device,
        batch_size=batch_size,
        preprocess_fn=preprocess_batch_fn,
        dtype=dtype,
    )


@torch.jit.script
def squared_pairwise_distance(
    x: torch.Tensor, y: torch.Tensor, a_min: float = 1e-30
) -> torch.Tensor:  # pragma: no cover - torch.jit.script code is compiled and copied
    """
    PyTorch pairwise squared Euclidean distance between samples x and y.

    Parameters
    ----------
    x
        Batch of instances of shape [Nx, features].
    y
        Batch of instances of shape [Ny, features].
    a_min
        Lower bound to clip distance values.
    Returns
    -------
    Pairwise squared Euclidean distance [Nx, Ny].
    """
    x2 = x.pow(2).sum(dim=-1, keepdim=True)
    y2 = y.pow(2).sum(dim=-1, keepdim=True)
    dist = torch.addmm(y2.transpose(-2, -1), x, y.transpose(-2, -1), alpha=-2).add_(x2)
    return dist.clamp_min_(a_min)


def sigma_median(x: torch.Tensor, y: torch.Tensor, dist: torch.Tensor) -> torch.Tensor:
    """
    Bandwidth estimation using the median heuristic :cite:t:`Gretton2012`.

    Parameters
    ----------
    x
        Tensor of instances with dimension [Nx, features].
    y
        Tensor of instances with dimension [Ny, features].
    dist
        Tensor with dimensions [Nx, Ny], containing the pairwise distances
        between `x` and `y`.

    Returns
    -------
    The computed bandwidth, `sigma`.
    """
    n = min(x.shape[0], y.shape[0])
    n = n if (x[:n] == y[:n]).all() and x.shape == y.shape else 0
    n_median = n + (np.prod(dist.shape) - n) // 2 - 1
    sigma = (0.5 * dist.flatten().sort().values[int(n_median)].unsqueeze(dim=-1)) ** 0.5
    return sigma


class GaussianRBF(nn.Module):
    """
    Gaussian RBF kernel: k(x,y) = exp(-(1/(2*sigma^2)||x-y||^2). A forward pass
    takes a batch of instances x [Nx, features] and y [Ny, features] and returns
    the kernel matrix [Nx, Ny].

    Parameters
    ----------
    sigma : Optional[torch.Tensor], default None
        Bandwidth used for the kernel. Needn't be specified if being inferred or
        trained. Can pass multiple values to eval kernel with and then average.
    init_sigma_fn : Optional[Callable], default None
        Function used to compute the bandwidth `sigma`. Used when `sigma` is to be
        inferred. The function's signature should take in the tensors `x`, `y` and
        `dist` and return `sigma`. If `None`, it is set to
        :func:`~daml._internal.detectors.drift.torch.sigma_median`.
    trainable : bool, default False
        Whether or not to track gradients w.r.t. `sigma` to allow it to be trained.
    """

    def __init__(
        self,
        sigma: Optional[torch.Tensor] = None,
        init_sigma_fn: Optional[Callable] = None,
        trainable: bool = False,
    ) -> None:
        super().__init__()
        init_sigma_fn = sigma_median if init_sigma_fn is None else init_sigma_fn
        self.config = {
            "sigma": sigma,
            "trainable": trainable,
            "init_sigma_fn": init_sigma_fn,
        }
        if sigma is None:
            self.log_sigma = nn.Parameter(torch.empty(1), requires_grad=trainable)
            self.init_required = True
        else:
            sigma = sigma.reshape(-1)  # [Ns,]
            self.log_sigma = nn.Parameter(sigma.log(), requires_grad=trainable)
            self.init_required = False
        self.init_sigma_fn = init_sigma_fn
        self.trainable = trainable

    @property
    def sigma(self) -> torch.Tensor:
        return self.log_sigma.exp()

    def forward(
        self,
        x: Union[np.ndarray, torch.Tensor],
        y: Union[np.ndarray, torch.Tensor],
        infer_sigma: bool = False,
    ) -> torch.Tensor:
        x, y = torch.as_tensor(x), torch.as_tensor(y)
        dist = squared_pairwise_distance(x.flatten(1), y.flatten(1))  # [Nx, Ny]

        if infer_sigma or self.init_required:
            if self.trainable and infer_sigma:
                raise ValueError("Gradients cannot be computed w.r.t. an inferred sigma value")
            sigma = self.init_sigma_fn(x, y, dist)
            with torch.no_grad():
                self.log_sigma.copy_(sigma.log().clone())
            self.init_required = False

        gamma = 1.0 / (2.0 * self.sigma**2)  # [Ns,]
        # TODO: do matrix multiplication after all?
        kernel_mat = torch.exp(-torch.cat([(g * dist)[None, :, :] for g in gamma], dim=0))  # [Ns, Nx, Ny]
        return kernel_mat.mean(dim=0)  # [Nx, Ny]



from enum import Flag
from typing import Any, Callable, Dict, Generic, Iterable, List, Optional, Sequence, TypeVar, Union

import numpy as np
from scipy.stats import entropy, kurtosis, skew

from daml._internal.flags import ImageHash, ImageProperty, ImageStatistics, ImageStatsFlags, ImageVisuals
from daml._internal.metrics.base import MetricMixin
from daml._internal.metrics.hash import pchash, xxhash
from daml._internal.metrics.utils import edge_filter, get_bitdepth, normalize_image_shape, rescale

QUARTILES = (0, 25, 50, 75, 100)

TBatch = TypeVar("TBatch", bound=Sequence)
TFlag = TypeVar("TFlag", bound=Flag)


class BaseStatsMetric(MetricMixin, Generic[TBatch, TFlag]):
    def __init__(self, flags: TFlag):
        self.flags = flags
        self.results = []

    def update(self, preds: TBatch, targets=None) -> None:
        """
        Updates internal metric cache for later calculation

        Parameters
        ----------
        batch : Sequence
            Sequence of images to be processed
        """

    def compute(self) -> Dict[str, Any]:
        """
        Computes the specified measures on the cached values

        Returns
        -------
        Dict[str, Any]
            Dictionary results of the specified measures
        """
        return {stat: [result[stat] for result in self.results] for stat in self.results[0]}

    def reset(self) -> None:
        """
        Resets the internal metric cache
        """
        self.results = []

    def _map(self, func_map: Dict[Flag, Callable]) -> Dict[str, Any]:
        """Calculates the measures for each flag if it is selected."""
        results = {}
        for flag, func in func_map.items():
            if not flag.name:
                raise ValueError("Provided flag to set value does not have a name.")
            if flag & self.flags:
                results[flag.name.lower()] = func()
        return results

    def _keys(self) -> List[str]:
        """Returns the list of measures to be calculated."""
        flags = (
            self.flags
            if isinstance(self.flags, Iterable)  # py3.11
            else [flag for flag in list(self.flags.__class__) if flag & self.flags]
        )
        return [flag.name.lower() for flag in flags if flag.name is not None]


class ImageHashMetric(BaseStatsMetric):
    """
    Hashes images using the specified algorithms

    Parameters
    ----------
    flags : ImageHash
        Algorithm(s) to calculate a hash as hex digest
    """

    def __init__(self, flags: ImageHash = ImageHash.ALL):
        super().__init__(flags)

    def update(self, preds: Iterable[np.ndarray], targets=None) -> None:
        for data in preds:
            results = self._map(
                {
                    ImageHash.XXHASH: lambda: xxhash(data),
                    ImageHash.PCHASH: lambda: pchash(data),
                }
            )
            self.results.append(results)


class ImagePropertyMetric(BaseStatsMetric):
    """
    Calculates specified image properties

    Parameters
    ----------
    flags: ImageProperty
        Property(ies) to calculate for each image
    """

    def __init__(self, flags: ImageProperty = ImageProperty.ALL):
        super().__init__(flags)

    def update(self, preds: Iterable[np.ndarray], targets=None) -> None:
        for data in preds:
            results = self._map(
                {
                    ImageProperty.WIDTH: lambda: np.int32(data.shape[-1]),
                    ImageProperty.HEIGHT: lambda: np.int32(data.shape[-2]),
                    ImageProperty.SIZE: lambda: np.int32(data.shape[-1] * data.shape[-2]),
                    ImageProperty.ASPECT_RATIO: lambda: data.shape[-1] / np.int32(data.shape[-2]),
                    ImageProperty.CHANNELS: lambda: data.shape[-3],
                    ImageProperty.DEPTH: lambda: get_bitdepth(data).depth,
                }
            )
            self.results.append(results)


class ImageVisualsMetric(BaseStatsMetric):
    """
    Calculates specified visual image properties

    Parameters
    ----------
    flags: ImageVisuals
        Property(ies) to calculate for each image
    """

    def __init__(self, flags: ImageVisuals = ImageVisuals.ALL):
        super().__init__(flags)

    def update(self, preds: Iterable[np.ndarray], targets=None) -> None:
        for data in preds:
            results = self._map(
                {
                    ImageVisuals.BRIGHTNESS: lambda: np.mean(rescale(data)),
                    ImageVisuals.BLURRINESS: lambda: np.std(edge_filter(np.mean(data, axis=0))),
                    ImageVisuals.MISSING: lambda: np.sum(np.isnan(data)),
                    ImageVisuals.ZERO: lambda: np.int32(np.count_nonzero(data == 0)),
                }
            )
            self.results.append(results)


class ImageStatisticsMetric(BaseStatsMetric):
    """
    Calculates descriptive statistics for each image

    Parameters
    ----------
    flags: ImageStatistics
        Statistic(s) to calculate for each image
    """

    def __init__(self, flags: ImageStatistics = ImageStatistics.ALL):
        super().__init__(flags)

    def update(self, preds: Iterable[np.ndarray], targets=None) -> None:
        for data in preds:
            scaled = rescale(data)
            if (ImageStatistics.HISTOGRAM | ImageStatistics.ENTROPY) & self.flags:
                hist = np.histogram(scaled, bins=256, range=(0, 1))[0]

            results = self._map(
                {
                    ImageStatistics.MEAN: lambda: np.mean(scaled),
                    ImageStatistics.STD: lambda: np.std(scaled),
                    ImageStatistics.VAR: lambda: np.var(scaled),
                    ImageStatistics.SKEW: lambda: np.float32(skew(scaled.ravel())),
                    ImageStatistics.KURTOSIS: lambda: np.float32(kurtosis(scaled.ravel())),
                    ImageStatistics.PERCENTILES: lambda: np.percentile(scaled, q=QUARTILES),
                    ImageStatistics.HISTOGRAM: lambda: hist,
                    ImageStatistics.ENTROPY: lambda: np.float32(entropy(hist)),
                }
            )
            self.results.append(results)


class ChannelStatisticsMetric(BaseStatsMetric):
    """
    Calculates descriptive statistics for each image per channel

    Parameters
    ----------
    flags: ImageStatistics
        Statistic(s) to calculate for each image per channel
    """

    def __init__(self, flags: ImageStatistics = ImageStatistics.ALL):
        super().__init__(flags)

    def update(self, preds: Iterable[np.ndarray], targets=None) -> None:
        for data in preds:
            scaled = rescale(data)
            flattened = scaled.reshape(data.shape[0], -1)

            if (ImageStatistics.HISTOGRAM | ImageStatistics.ENTROPY) & self.flags:
                hist = np.apply_along_axis(lambda x: np.histogram(x, bins=256, range=(0, 1))[0], 1, flattened)

            results = self._map(
                {
                    ImageStatistics.MEAN: lambda: np.mean(flattened, axis=1),
                    ImageStatistics.STD: lambda: np.std(flattened, axis=1),
                    ImageStatistics.VAR: lambda: np.var(flattened, axis=1),
                    ImageStatistics.SKEW: lambda: skew(flattened, axis=1),
                    ImageStatistics.KURTOSIS: lambda: kurtosis(flattened, axis=1),
                    ImageStatistics.PERCENTILES: lambda: np.percentile(flattened, q=QUARTILES, axis=1).T,
                    ImageStatistics.HISTOGRAM: lambda: hist,
                    ImageStatistics.ENTROPY: lambda: entropy(hist, axis=1),
                }
            )
            self.results.append(results)


class BaseAggregateMetric(BaseStatsMetric, Generic[TFlag]):
    FLAG_METRIC_MAP: Dict[type, type]
    DEFAULT_FLAGS: Sequence[TFlag]

    def __init__(self, flags: Optional[Union[TFlag, Sequence[TFlag]]] = None):
        flag_dict = {}
        for flag in flags if isinstance(flags, Sequence) else self.DEFAULT_FLAGS if not flags else [flags]:
            flag_dict[type(flag)] = flag_dict.setdefault(type(flag), type(flag)(0)) | flag
        self._metrics_dict = {
            metric: []
            for metric in (
                self.FLAG_METRIC_MAP[flag_class](flag) for flag_class, flag in flag_dict.items() if flag.value != 0
            )
        }


class ImageStats(BaseAggregateMetric):
    """
    Calculates various image property statistics

    Parameters
    ----------
    flags: [ImageHash | ImageProperty | ImageStatistics | ImageVisuals], default None
        Metric(s) to calculate for each image per channel - calculates all metrics if None
    """

    FLAG_METRIC_MAP = {
        ImageHash: ImageHashMetric,
        ImageProperty: ImagePropertyMetric,
        ImageStatistics: ImageStatisticsMetric,
        ImageVisuals: ImageVisualsMetric,
    }
    DEFAULT_FLAGS = [ImageHash.ALL, ImageProperty.ALL, ImageStatistics.ALL, ImageVisuals.ALL]

    def __init__(self, flags: Optional[Union[ImageStatsFlags, Sequence[ImageStatsFlags]]] = None):
        super().__init__(flags)
        self._length = 0

    def update(self, preds: Iterable[np.ndarray], targets=None) -> None:
        for image in preds:
            self._length += 1
            img = normalize_image_shape(image)
            for metric in self._metrics_dict:
                metric.update([img])

    def compute(self) -> Dict[str, Any]:
        for metric in self._metrics_dict:
            self._metrics_dict[metric] = metric.results

        stats = {}
        for metric, results in self._metrics_dict.items():
            for i, result in enumerate(results):
                for stat in metric._keys():
                    value = result[stat]
                    if not isinstance(value, (np.ndarray, np.generic)):
                        if stat not in stats:
                            stats[stat] = []
                        stats[stat].append(result[stat])
                    else:
                        if stat not in stats:
                            shape = () if np.isscalar(result[stat]) else result[stat].shape
                            stats[stat] = np.empty((self._length,) + shape)
                        stats[stat][i] = result[stat]
        return stats

    def reset(self):
        self._length = 0
        for metric in self._metrics_dict:
            metric.reset()
            self._metrics_dict[metric] = []


class ChannelStats(BaseAggregateMetric):
    FLAG_METRIC_MAP = {ImageStatistics: ChannelStatisticsMetric}
    DEFAULT_FLAGS = [ImageStatistics.ALL]
    IDX_MAP = "idx_map"

    def __init__(self, flags: Optional[ImageStatistics] = None) -> None:
        super().__init__(flags)

    def update(self, preds: Iterable[np.ndarray], targets=None) -> None:
        for image in preds:
            img = normalize_image_shape(image)
            for metric in self._metrics_dict:
                metric.update([img])

        for metric in self._metrics_dict:
            self._metrics_dict[metric] = metric.results

    def compute(self) -> Dict[str, Any]:
        # Aggregate all metrics into a single dictionary
        stats = {}
        channel_stats = set()
        for metric, results in self._metrics_dict.items():
            for i, result in enumerate(results):
                for stat in metric._keys():
                    channel_stats.update(metric._keys())
                    channels = result[stat].shape[0]
                    stats.setdefault(self.IDX_MAP, {}).setdefault(channels, {})[i] = None
                    stats.setdefault(stat, {}).setdefault(channels, []).append(result[stat])

        # Concatenate list of channel statistics numpy
        for stat in channel_stats:
            for channel in stats[stat]:
                stats[stat][channel] = np.array(stats[stat][channel]).T

        for channel in stats[self.IDX_MAP]:
            stats[self.IDX_MAP][channel] = list(stats[self.IDX_MAP][channel].keys())

        return stats

    def reset(self) -> None:
        for metric in self._metrics_dict:
            metric.reset()
            self._metrics_dict[metric] = []



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from typing import Callable

import numpy as np
from keras.api._v2 import keras

from daml._internal.detectors.ood.base import OODGMMBase, OODScore
from daml._internal.models.tensorflow.autoencoder import VAEGMM
from daml._internal.models.tensorflow.gmm import gmm_energy
from daml._internal.models.tensorflow.losses import Elbo, LossGMM
from daml._internal.models.tensorflow.utils import predict_batch


class OOD_VAEGMM(OODGMMBase):
    def __init__(self, model: VAEGMM, samples: int = 10) -> None:
        """
        VAE with Gaussian Mixture Model based outlier detector.

        Parameters
        ----------
        model : VAEGMM
            A VAEGMM model.
        samples
            Number of samples sampled to evaluate each instance.
        """
        super().__init__(model)
        self.samples = samples

    def fit(
        self,
        x_ref: np.ndarray,
        threshold_perc: float = 100.0,
        loss_fn: Callable = LossGMM(elbo=Elbo(0.05)),
        optimizer: keras.optimizers.Optimizer = keras.optimizers.Adam,
        epochs: int = 20,
        batch_size: int = 64,
        verbose: bool = True,
    ) -> None:
        """
        Train the AE model with recommended loss function and optimizer.

        Parameters
        ----------
        X : np.ndarray
            Training batch.
        threshold_perc : float, default 100.0
            Percentage of reference data that is normal.
        loss_fn : Callable, default LossGMM(elbo=Elbo(0.05))
            Loss function used for training.
        optimizer : keras.optimizers.Optimizer, default keras.optimizers.Adam
            Optimizer used for training.
        epochs : int, default 20
            Number of training epochs.
        batch_size : int, default 64
            Batch size used for training.
        verbose : bool, default True
            Whether to print training progress.
        """
        super().fit(x_ref, threshold_perc, loss_fn, optimizer, epochs, batch_size, verbose)

    def score(self, X: np.ndarray, batch_size: int = int(1e10)) -> OODScore:
        self._validate(X)

        # draw samples from latent space
        X_samples = np.repeat(X, self.samples, axis=0)
        _, z, _ = predict_batch(X_samples, self.model, batch_size=batch_size)

        # compute average energy for samples
        energy, _ = gmm_energy(z, self.gmm_params, return_mean=False)
        energy_samples = energy.numpy().reshape((-1, self.samples))  # type: ignore
        iscore = np.mean(energy_samples, axis=-1)
        return OODScore(iscore)



import numpy as np
import xxhash as xxh
from PIL import Image
from scipy.fftpack import dct

from daml._internal.metrics.utils import normalize_image_shape, rescale

HASH_SIZE = 8
MAX_FACTOR = 4


def pchash(image: np.ndarray) -> str:
    """
    Performs a perceptual hash on an image by resizing to a square NxN image
    using the Lanczos algorithm where N is 32x32 or the largest multiple of
    8 that is smaller than the input image dimensions.  The resampled image
    is compressed using a discrete cosine transform and the lowest frequency
    component is encoded as a bit array of greater or less than median value
    and returned as a hex string.

    Parameters
    ----------
    image : np.ndarray
        An image as a numpy array in CxHxW format

    Returns
    -------
    str
        The hex string hash of the image using perceptual hashing
    """
    # Verify that the image is at least larger than an 8x8 image
    min_dim = min(image.shape[-2:])
    if min_dim < HASH_SIZE + 1:
        raise ValueError(f"Image must be larger than {HASH_SIZE}x{HASH_SIZE} for fuzzy hashing.")

    # Calculates the dimensions of the resized square image
    resize_dim = HASH_SIZE * min((min_dim - 1) // HASH_SIZE, MAX_FACTOR)

    # Normalizes the image to CxHxW and takes the mean over all the channels
    normalized = np.mean(normalize_image_shape(image), axis=0).squeeze()

    # Rescales the pixel values to an 8-bit 0-255 image
    rescaled = rescale(normalized, 8).astype(np.uint8)

    # Resizes the image using the Lanczos algorithm to a square image
    im = np.array(Image.fromarray(rescaled).resize((resize_dim, resize_dim), Image.Resampling.LANCZOS))

    # Performs discrete cosine transforms to compress the image information and takes the lowest frequency component
    transform = dct(dct(im.T).T)[:HASH_SIZE, :HASH_SIZE]

    # Encodes the transform as a bit array over the median value
    diff = transform > np.median(transform)

    # Pads the front of the bit array to a multiple of 8 with False
    padded = np.full(int(np.ceil(diff.size / 8) * 8), False)
    padded[-diff.size :] = diff.ravel()

    # Converts the bit array to a hex string and strips leading 0s
    hash_hex = np.packbits(padded).tobytes().hex().lstrip("0")
    return hash_hex if hash_hex else "0"


def xxhash(image: np.ndarray) -> str:
    """
    Performs a fast non-cryptographic hash using the xxhash algorithm
    (xxhash.com) against the image as a flattened bytearray.  The hash
    is returned as a hex string.

    Parameters
    ----------
    image : np.ndarray
        An image as a numpy array

    Returns
    -------
    str
        The hex string hash of the image using the xxHash algorithm
    """
    return xxh.xxh3_64_hexdigest(image.ravel().tobytes())



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from typing import Literal, Optional, Union, cast

import numpy as np
import tensorflow as tf
from keras.api._v2.keras.layers import Flatten
from tensorflow_probability.python.distributions.mvn_diag import MultivariateNormalDiag
from tensorflow_probability.python.distributions.mvn_tril import MultivariateNormalTriL
from tensorflow_probability.python.stats import covariance

from daml._internal.models.tensorflow.gmm import gmm_energy, gmm_params


class Elbo:
    """
    Compute ELBO loss. The covariance matrix can be specified by passing the full covariance matrix, the matrix
    diagonal, or a scale identity multiplier. Only one of these should be specified. If none are specified, the
    identity matrix is used.

    Parameters
    ----------
    cov_type
        Full covariance matrix, diagonal variance matrix, or scale identity multiplier.
    x
        Dataset used to calculate the covariance matrix.  Required for full and diagonal covariance matrix types.
    """

    def __init__(
        self,
        cov_type: Union[Literal["cov_full", "cov_diag"], float] = 1.0,
        x: Optional[Union[tf.Tensor, np.ndarray]] = None,
    ):
        if isinstance(cov_type, float):
            self.cov = ("sim", cov_type)
        elif cov_type in ["cov_full", "cov_diag"]:
            x_np: np.ndarray = x.numpy() if tf.is_tensor(x) else x  # type: ignore
            cov = covariance(x_np.reshape(x_np.shape[0], -1))  # type: ignore py38
            if cov_type == "cov_diag":  # infer standard deviation from covariance matrix
                cov = tf.math.sqrt(tf.linalg.diag_part(cov))
            self.cov = (cov_type, cov)
        else:
            raise ValueError("Only cov_full, cov_diag or sim value should be specified.")

    def __call__(self, y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
        y_pred_flat = cast(tf.Tensor, Flatten()(y_pred))

        if self.cov[0] == "cov_full":
            y_mn = MultivariateNormalTriL(y_pred_flat, scale_tril=tf.linalg.cholesky(self.cov[1]))
        else:  # cov_diag and sim
            cov_diag = self.cov[1] if self.cov[0] == "cov_diag" else self.cov[1] * tf.ones(y_pred_flat.shape[-1])
            y_mn = MultivariateNormalDiag(y_pred_flat, scale_diag=cov_diag)

        loss = -tf.reduce_mean(y_mn.log_prob(Flatten()(y_true)))
        return loss


class LossGMM:
    """
    Loss function used for AE and VAE with GMM.

    Parameters
    ----------
    w_recon
        Weight on elbo loss term.
    w_energy
        Weight on sample energy loss term.
    w_cov_diag
        Weight on covariance regularizing loss term.
    elbo
        ELBO loss function used to calculate w_recon.
    """

    def __init__(
        self,
        w_recon: float = 1e-7,
        w_energy: float = 0.1,
        w_cov_diag: float = 0.005,
        elbo: Optional[Elbo] = None,
    ):
        self.w_recon = w_recon
        self.w_energy = w_energy
        self.w_cov_diag = w_cov_diag
        self.elbo = elbo

    def __call__(
        self,
        x_true: tf.Tensor,
        x_pred: tf.Tensor,
        z: tf.Tensor,
        gamma: tf.Tensor,
    ) -> tf.Tensor:
        w_recon = (
            tf.reduce_mean(tf.subtract(x_true, x_pred) ** 2)
            if self.elbo is None
            else tf.multiply(self.w_recon, self.elbo(x_true, x_pred))
        )
        sample_energy, cov_diag = gmm_energy(z, gmm_params(z, gamma))
        w_energy = tf.multiply(self.w_energy, sample_energy)
        w_cov_diag = tf.multiply(self.w_cov_diag, cov_diag)
        return w_recon + w_energy + w_cov_diag



from numpy import float32, ndarray
from torch import Tensor, from_numpy


def torch_to_numpy(tensor: Tensor) -> ndarray:
    """
    Converts a PyTorch tensor to a NumPy array
    """
    if isinstance(tensor, ndarray):  # Already array, return
        return tensor
    if not isinstance(tensor, Tensor):
        raise TypeError("Tensor is not of type Tensor")

    x: ndarray = tensor.detach().cpu().numpy()
    return x


def numpy_to_torch(array: ndarray) -> Tensor:
    """
    Converts a NumPy array to a PyTorch tensor
    """
    if isinstance(array, Tensor):  # Already tensor, return
        return array
    if not isinstance(array, ndarray):
        raise TypeError("Array is not of type numpy.ndarray")
    x: Tensor = from_numpy(array.astype(float32))
    return x


def permute_to_torch(array: ndarray) -> Tensor:
    """
    Converts and permutes a NumPy image array into a PyTorch image tensor.

    Parameters
    ----------
    array: ndarray
        Array containing image data in the format NHWC

    Returns
    -------
    Tensor
        Tensor containing image data in the format NCHW
    """
    x = numpy_to_torch(array)
    x = x.permute(0, 3, 1, 2)  # NHWC -> NCHW
    return x


def permute_to_numpy(tensor: Tensor) -> ndarray:
    """
    Converts and permutes a PyTorch image tensor into a NumPy image array.

    Does not permute if given ndarray

    Parameters
    ----------
    tensor: Tensor
        Tensor containing image data in the format NCHW

    Returns
    -------
    ndarray
        Array containing image data in the format NHWC
    """
    x = tensor.permute(0, 2, 3, 1)
    x = torch_to_numpy(x)  # NCHW -> NHWC
    return x



from typing import Literal, Optional, Sequence, Union

import numpy as np

from daml._internal.flags import ImageProperty, ImageVisuals, LinterFlags
from daml._internal.metrics.stats import ImageStats


def _get_outlier_mask(
    values: np.ndarray, method: Literal["zscore", "modzscore", "iqr"], threshold: Optional[float]
) -> np.ndarray:
    if method == "zscore":
        threshold = threshold if threshold else 3.0
        std = np.std(values)
        abs_diff = np.abs(values - np.mean(values))
        return (abs_diff / std) > threshold
    elif method == "modzscore":
        threshold = threshold if threshold else 3.5
        abs_diff = np.abs(values - np.mean(values))
        med_abs_diff = np.median(abs_diff)
        mod_z_score = 0.6745 * abs_diff / med_abs_diff
        return mod_z_score > threshold
    elif method == "iqr":
        threshold = threshold if threshold else 1.5
        qrt = np.percentile(values, q=(25, 75), method="midpoint")
        iqr = (qrt[1] - qrt[0]) * threshold
        return (values < (qrt[0] - iqr)) | (values > (qrt[1] + iqr))
    else:
        raise ValueError("Outlier method must be 'zscore' 'modzscore' or 'iqr'.")


class Linter:
    """
    Calculates statistical outliers of a dataset using various statistical
    tests applied to each image
    """

    def __init__(
        self,
        images: np.ndarray,
        flags: Optional[Union[LinterFlags, Sequence[LinterFlags]]] = None,
    ):
        flags = flags if flags is not None else (ImageProperty.ALL, ImageVisuals.ALL)
        self.stats = ImageStats(flags)
        self.images = images

    def _get_outliers(
        self,
        outlier_method: Literal["zscore", "modzscore", "iqr"] = "modzscore",
        outlier_threshold: Optional[float] = None,
    ) -> dict:
        flagged_images = {}

        for stat, values in self.results.items():
            if not isinstance(values, np.ndarray):
                continue

            if values.ndim == 1 and np.std(values) != 0:
                mask = _get_outlier_mask(values, outlier_method, outlier_threshold)
                indices = np.flatnonzero(mask)
                for i, value in zip(indices, values[mask]):
                    flagged_images.setdefault(i, {}).update({stat: np.round(value, 2)})

        return dict(sorted(flagged_images.items()))

    def evaluate(self) -> dict:
        """
        Returns indices of outliers with and the issues identified for each

        Returns
        -------
        Dict[int, Dict[str, float]]
            Dictionary containing the indices of outliers and a dictionary issues and calculated values
        """
        self.stats.reset()
        self.stats.update(self.images)
        self.results = self.stats.compute()
        return self._get_outliers()



"""
This module contains the implementation of the
FR Test Statistic based estimate for the upperbound
average precision using empirical mean precision
"""

from typing import Dict

import numpy as np
from sklearn.metrics import average_precision_score

from daml._internal.metrics.base import EvaluateMixin


class UAP(EvaluateMixin):
    """
    FR Test Statistic based estimate of the empirical mean precision

    Parameters
    ----------
    labels : np.ndarray
        A numpy array of n_samples of class labels with M unique classes.

    scores : np.ndarray
        A 2D array of class probabilities per image
    """

    def __init__(self, labels: np.ndarray, scores: np.ndarray) -> None:
        self.labels = labels
        self.scores = scores

    def evaluate(self) -> Dict[str, float]:
        """
        Returns
        -------
        Dict[str, float]
            uap : The empirical mean precision estimate

        Raises
        ------
        ValueError
            If unique classes M < 2
        """
        uap = float(average_precision_score(self.labels, self.scores, average="weighted"))
        return {"uap": uap}



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from abc import ABC, abstractmethod
from functools import wraps
from random import random
from typing import Callable, Dict, Literal, Optional, Tuple, Union

import numpy as np


def update_x_ref(fn):
    @wraps(fn)
    def _(self, x, *args, **kwargs):
        output = fn(self, x, *args, **kwargs)

        # update reference dataset
        if self.update_x_ref is not None:
            self._x_ref = self.update_x_ref(self.x_ref, x, self.n)

        # used for reservoir sampling
        self.n += len(x)
        return output

    return _


def preprocess_x(fn):
    @wraps(fn)
    def _(self, x, *args, **kwargs):
        if self._x_refcount == 0:
            self._x = self._preprocess(x)
        self._x_refcount += 1
        output = fn(self, self._x, *args, **kwargs)
        self._x_refcount -= 1
        if self._x_refcount == 0:
            del self._x
        return output

    return _


class UpdateStrategy(ABC):
    def __init__(self, n: int):
        self.n = n

    @abstractmethod
    def __call__(self, x_ref: np.ndarray, x: np.ndarray, count: int) -> np.ndarray:
        """Abstract implementation of update strategy"""


class LastSeenUpdate(UpdateStrategy):
    """
    Updates reference dataset for drift detector using last seen method.

    Parameters
    ----------
    n : int
        Update with last n instances seen by the detector.
    """

    def __call__(self, x_ref: np.ndarray, x: np.ndarray, count: int) -> np.ndarray:
        x_updated = np.concatenate([x_ref, x], axis=0)
        return x_updated[-self.n :]


class ReservoirSamplingUpdate(UpdateStrategy):
    """
    Updates reference dataset for drift detector using reservoir sampling method.

    Parameters
    ----------
    n : int
        Update with reservoir sampling of size n.
    """

    def __call__(self, x_ref: np.ndarray, x: np.ndarray, count: int) -> np.ndarray:
        if x.shape[0] + count <= self.n:
            return np.concatenate([x_ref, x], axis=0)

        n_ref = x_ref.shape[0]
        output_size = min(self.n, n_ref + x.shape[0])
        shape = (output_size,) + x.shape[1:]
        x_reservoir = np.zeros(shape, dtype=x_ref.dtype)
        x_reservoir[:n_ref] = x_ref
        for item in x:
            count += 1
            if n_ref < self.n:
                x_reservoir[n_ref, :] = item
                n_ref += 1
            else:
                r = int(random() * count)
                if r < self.n:
                    x_reservoir[r, :] = item
        return x_reservoir


class BaseDrift:
    """Generic drift detector component handling preprocessing of data and correction"""

    def __init__(
        self,
        x_ref: np.ndarray,
        p_val: float = 0.05,
        x_ref_preprocessed: bool = False,
        update_x_ref: Optional[UpdateStrategy] = None,
        preprocess_fn: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        correction: Literal["bonferroni", "fdr"] = "bonferroni",
    ) -> None:
        # Type checking
        if preprocess_fn is not None and not isinstance(preprocess_fn, Callable):
            raise ValueError("`preprocess_fn` is not a valid Callable.")
        if update_x_ref is not None and not isinstance(update_x_ref, UpdateStrategy):
            raise ValueError("`update_x_ref` is not a valid ReferenceUpdate class.")
        if correction not in ["bonferroni", "fdr"]:
            raise ValueError("`correction` must be `bonferroni` or `fdr`.")

        self._x_ref = x_ref
        self.x_ref_preprocessed = x_ref_preprocessed

        # Other attributes
        self.p_val = p_val
        self.update_x_ref = update_x_ref
        self.preprocess_fn = preprocess_fn
        self.correction = correction
        self.n = len(x_ref)

        # Ref counter for preprocessed x
        self._x_refcount = 0

    @property
    def x_ref(self) -> np.ndarray:
        if not self.x_ref_preprocessed:
            self.x_ref_preprocessed = True
            if self.preprocess_fn is not None:
                self._x_ref = self.preprocess_fn(self._x_ref)

        return self._x_ref

    def _preprocess(self, x: np.ndarray) -> np.ndarray:
        """Data preprocessing before computing the drift scores."""
        if self.preprocess_fn is not None:
            x = self.preprocess_fn(x)
        return x


class BaseUnivariateDrift(BaseDrift):
    """
    Generic drift detector component which serves as a base class for methods using
    univariate tests. If n_features > 1, a multivariate correction is applied such
    that the false positive rate is upper bounded by the specified p-value, with
    equality in the case of independent features.
    """

    def __init__(
        self,
        x_ref: np.ndarray,
        p_val: float = 0.05,
        x_ref_preprocessed: bool = False,
        update_x_ref: Optional[UpdateStrategy] = None,
        preprocess_fn: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        correction: Literal["bonferroni", "fdr"] = "bonferroni",
        n_features: Optional[int] = None,
    ) -> None:
        super().__init__(
            x_ref,
            p_val,
            x_ref_preprocessed,
            update_x_ref,
            preprocess_fn,
            correction,
        )

        self._n_features = n_features

    @property
    def n_features(self) -> int:
        # lazy process n_features as needed
        if not isinstance(self._n_features, int):
            # compute number of features for the univariate tests
            if not isinstance(self.preprocess_fn, Callable) or self.x_ref_preprocessed:
                # infer features from preprocessed reference data
                self._n_features = self.x_ref.reshape(self.x_ref.shape[0], -1).shape[-1]
            else:
                # infer number of features after applying preprocessing step
                x = self.preprocess_fn(self.x_ref[0:1])
                self._n_features = x.reshape(x.shape[0], -1).shape[-1]

        return self._n_features

    @preprocess_x
    @abstractmethod
    def score(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """Abstract method to calculate feature score after preprocessing"""

    def _apply_correction(self, p_vals: np.ndarray) -> Tuple[int, float]:
        if self.correction == "bonferroni":
            threshold = self.p_val / self.n_features
            drift_pred = int((p_vals < threshold).any())
            return drift_pred, threshold
        elif self.correction == "fdr":
            n = p_vals.shape[0]
            i = np.arange(n) + 1
            p_sorted = np.sort(p_vals)
            q_threshold = self.p_val * i / n
            below_threshold = p_sorted < q_threshold
            try:
                idx_threshold = int(np.where(below_threshold)[0].max())
            except ValueError:  # sorted p-values not below thresholds
                return int(below_threshold.any()), q_threshold.min()
            return int(below_threshold.any()), q_threshold[idx_threshold]
        else:
            raise ValueError("`correction` needs to be either `bonferroni` or `fdr`.")

    @preprocess_x
    @update_x_ref
    def predict(
        self,
        x: np.ndarray,
        drift_type: Literal["batch", "feature"] = "batch",
    ) -> Dict[str, Union[int, float, np.ndarray]]:
        """
        Predict whether a batch of data has drifted from the reference data and update
        reference data using specified update strategy.

        Parameters
        ----------
        x : np.ndarray
            Batch of instances.
        drift_type : Literal["batch", "feature"], default "batch"
            Predict drift at the 'feature' or 'batch' level. For 'batch', the test
            statistics for each feature are aggregated using the Bonferroni or False
            Discovery Rate correction (if n_features>1).

        Returns
        -------
        Dictionary containing the drift prediction and optionally the feature level
                p-values, threshold after multivariate correction if needed and test
                statistics.
        """
        # compute drift scores
        p_vals, dist = self.score(x)

        # TODO: return both feature-level and batch-level drift predictions by default
        # values below p-value threshold are drift
        if drift_type == "feature":
            drift_pred = (p_vals < self.p_val).astype(int)
            threshold = self.p_val
        elif drift_type == "batch":
            drift_pred, threshold = self._apply_correction(p_vals)
        else:
            raise ValueError("`drift_type` needs to be either `feature` or `batch`.")

        # populate drift dict
        return {
            "is_drift": drift_pred,
            "p_val": p_vals,
            "threshold": threshold,
            "distance": dist,
        }



import math
from typing import Optional, Tuple

import numpy as np
from scipy.spatial.distance import pdist, squareform


class Coverage:
    """
     Class for evaluating coverage and identifying images/samples which are in undercovered regions.

     Idea behind this implementation comes from https://dl.acm.org/doi/abs/10.1145/3448016.3457315.

     Parameters
     ----------
     embeddings : np.ndarray
         n x p array of image embeddings from the dataset (reduced in dimension with an autoencoder or similar).
        note: Embeddings should be on the unit interval.
     radius_type : Optional[str]
         Either "naive" or "adaptive", denoting the type of radius to consider for coverage (default = adaptive).
    k: int
        Number of observations required in order to be covered. Default is 20 (based on the above paper and
          the central limit theorem).
    percent: np.float64
         Percent of observations to be considered uncovered. Only applies to adaptive radius (default is 0.01).

    Raises
    ------
    ValueError
         If radius type is not one of the two accepted values.
    """

    def __init__(
        self,
        embeddings: np.ndarray,
        radius_type: Optional[str] = "adaptive",
        k: int = 20,
        percent: np.float64 = np.float64(0.01),
    ):
        self.embeddings = embeddings
        self.radius_type = radius_type
        self.k = k
        self.percent = percent

    def evaluate(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Perform a one-way chi-squared test between observation frequencies and expected frequencies that
        tests the null hypothesis that the observed data has the expected frequencies.

        This function acts as an interface to the scipy.stats.chisquare method, which is documented at
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.chisquare.html

        Returns
        -------
        np.ndarray
            Array of uncovered indices
        np.ndarray
            Array of critical value radii
        """

        # Calculate distance matrix, look at the (k+1)st farthest neighbor for each image.
        n = len(self.embeddings)
        if n <= self.k:
            raise ValueError("Number of observations less than specified number of neighbors.")
        mat = squareform(pdist(self.embeddings))
        sorted_dists = np.sort(mat, axis=1)
        crit = sorted_dists[:, self.k + 1]

        d = np.shape(self.embeddings)[1]
        if self.radius_type == "naive":
            self.rho = (1 / math.sqrt(math.pi)) * ((2 * self.k * math.gamma(d / 2 + 1)) / (n)) ** (1 / d)
            pvals = np.where(crit > self.rho)[0]
        elif self.radius_type == "adaptive":
            # Use data adaptive cutoff
            cutoff = int(n * self.percent)
            pvals = np.argsort(crit)[::-1][:cutoff]
        else:
            raise ValueError("Invalid radius type.")
        return pvals, crit



from importlib.util import find_spec

from . import flags, metrics

__version__ = "0.55.0"

__all__ = ["metrics", "flags"]

if find_spec("torch") is not None:  # pragma: no cover
    from . import models, workflows

    __all__ += ["models", "workflows"]
elif find_spec("tensorflow") is not None:  # pragma: no cover
    from . import models

    __all__ += ["models"]

del find_spec



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from functools import partial
from typing import Callable, Optional, Tuple

import numpy as np
import tensorflow as tf
from keras.api._v2 import keras
from keras.api._v2.keras.layers import Input
from keras.api._v2.keras.models import Model

from daml._internal.detectors.ood.base import OODBase, OODScore
from daml._internal.models.tensorflow.pixelcnn import PixelCNN
from daml._internal.models.tensorflow.trainer import trainer
from daml._internal.models.tensorflow.utils import predict_batch


def build_model(
    dist: PixelCNN, input_shape: Optional[tuple] = None, filepath: Optional[str] = None
) -> Tuple[keras.Model, PixelCNN]:
    """
    Create keras.Model from TF distribution.

    Parameters
    ----------
    dist
        TensorFlow distribution.
    input_shape
        Input shape of the model.
    filepath
        File to load model weights from.

    Returns
    -------
    TensorFlow model.
    """
    x_in = Input(shape=input_shape)
    log_prob = dist.log_prob(x_in)
    model = Model(inputs=x_in, outputs=log_prob)
    model.add_loss(-tf.reduce_mean(log_prob))
    if isinstance(filepath, str):
        model.load_weights(filepath)
    return model, dist


def mutate_categorical(
    X: np.ndarray,
    rate: float,
    seed: int = 0,
    feature_range: tuple = (0, 255),
) -> tf.Tensor:
    """
    Randomly change integer feature values to values within a set range
    with a specified permutation rate.

    Parameters
    ----------
    X
        Batch of data to be perturbed.
    rate
        Permutation rate (between 0 and 1).
    seed
        Random seed.
    feature_range
        Min and max range for perturbed features.

    Returns
    -------
    Array with perturbed data.
    """
    frange = (feature_range[0] + 1, feature_range[1] + 1)
    shape = X.shape
    n_samples = np.prod(shape)
    mask = tf.random.categorical(tf.math.log([[1.0 - rate, rate]]), n_samples, seed=seed, dtype=tf.int32)
    mask = tf.reshape(mask, shape)
    possible_mutations = tf.random.uniform(shape, minval=frange[0], maxval=frange[1], dtype=tf.int32, seed=seed + 1)
    X = tf.math.floormod(tf.cast(X, tf.int32) + mask * possible_mutations, frange[1])  # type: ignore py38
    return tf.cast(X, tf.float32)  # type: ignore


class OOD_LLR(OODBase):
    def __init__(
        self,
        model: PixelCNN,
        model_background: Optional[PixelCNN] = None,
        log_prob: Optional[Callable] = None,
        sequential: bool = False,
    ) -> None:
        """
        Likelihood Ratios based outlier detector.

        Parameters
        ----------
        model : PixelCNN
            Generative distribution model.
        model_background : Optional[PixelCNN], default None
            Optional model for the background. Only needed if it is different from `model`.
        log_prob : Optional[Callable], default None
            Function used to evaluate log probabilities under the model
            if the model does not have a `log_prob` function.
        sequential : bool, default False
            Whether the data is sequential. Used to create targets during training.
        """
        self.dist_s = model
        self.dist_b = (
            model.copy()
            if hasattr(model, "copy")
            else keras.models.clone_model(model)
            if model_background is None
            else model_background
        )
        self.has_log_prob = hasattr(model, "log_prob")
        self.sequential = sequential
        self.log_prob = log_prob

        self._ref_score: OODScore
        self._threshold_perc: float
        self._data_info: Optional[Tuple[tuple, type]] = None

    def fit(
        self,
        x_ref: np.ndarray,
        threshold_perc: float = 100.0,
        loss_fn: Optional[Callable] = None,
        optimizer: keras.optimizers.Optimizer = keras.optimizers.Adam,
        epochs: int = 20,
        batch_size: int = 64,
        verbose: bool = True,
        mutate_fn: Callable = mutate_categorical,
        mutate_fn_kwargs: dict = {"rate": 0.2, "seed": 0, "feature_range": (0, 255)},
        mutate_batch_size: int = int(1e10),
    ) -> None:
        """
        Train semantic and background generative models.

        Parameters
        ----------
        x_ref : np.ndarray
            Training batch.
        threshold_perc : float, default 100.0
            Percentage of reference data that is normal.
        loss_fn : Optional[Callable], default None
            Loss function used for training.
        optimizer : keras.optimizers.Optimizer, default keras.optimizers.Adam
            Optimizer used for training.
        epochs : int, default 20
            Number of training epochs.
        batch_size : int, default 64
            Batch size used for training.
        verbose : bool, default True
            Whether to print training progress.
        mutate_fn : Callable, default mutate_categorical
            Mutation function used to generate the background dataset.
        mutate_fn_kwargs : dict, default {"rate": 0.2, "seed": 0, "feature_range": (0, 255)}
            Kwargs for the mutation function used to generate the background dataset.
            Default values set for an image dataset.
        mutate_batch_size: int, default int(1e10)
            Batch size used to generate the mutations for the background dataset.
        """
        input_shape = x_ref.shape[1:]
        optimizer = optimizer() if isinstance(optimizer, type) else optimizer
        # Separate into two separate optimizers, one for semantic model and one for background model
        optimizer_s = optimizer
        optimizer_b = optimizer.__class__.from_config(optimizer.get_config())

        # training arguments
        kwargs = {
            "epochs": epochs,
            "batch_size": batch_size,
            "verbose": verbose,
        }

        # create background data
        mutate_fn = partial(mutate_fn, **mutate_fn_kwargs)
        X_back = predict_batch(x_ref, mutate_fn, batch_size=mutate_batch_size, dtype=x_ref.dtype)

        # prepare sequential data
        if self.sequential and not self.has_log_prob:
            y, y_back = x_ref[:, 1:], X_back[:, 1:]  # type: ignore
            X, X_back = x_ref[:, :-1], X_back[:, :-1]  # type: ignore
        else:
            X = x_ref
            y, y_back = None, None

        # check if model needs to be built
        use_build = self.has_log_prob and not isinstance(self.dist_s, keras.Model)

        if use_build:
            # build and train semantic model
            self.model_s = build_model(self.dist_s, input_shape)[0]
            self.model_s.compile(optimizer=optimizer_s)
            self.model_s.fit(X, **kwargs)
            # build and train background model
            self.model_b = build_model(self.dist_b, input_shape)[0]
            self.model_b.compile(optimizer=optimizer_b)
            self.model_b.fit(X_back, **kwargs)
        else:
            # train semantic model
            args = [self.dist_s, X]
            kwargs.update({"y_train": y, "loss_fn": loss_fn, "optimizer": optimizer_s})
            trainer(*args, **kwargs)

            # train background model
            args = [self.dist_b, X_back]
            kwargs.update({"y_train": y_back, "loss_fn": loss_fn, "optimizer": optimizer_b})
            trainer(*args, **kwargs)

        self._datainfo = self._get_data_info(x_ref)
        self._ref_score = self.score(x_ref, batch_size=batch_size)
        self._threshold_perc = threshold_perc

    def _logp(
        self,
        dist,
        X: np.ndarray,
        return_per_feature: bool = False,
        batch_size: int = int(1e10),
    ) -> np.ndarray:
        """
        Compute log probability of a batch of instances under the generative model.
        """
        logp_fn = partial(dist.log_prob, return_per_feature=return_per_feature)
        # TODO: TBD: can this be any of the other types from predict_batch? i.e. tf.Tensor or tuple
        return predict_batch(X, logp_fn, batch_size=batch_size)  # type: ignore[return-value]

    def _logp_alt(
        self,
        model: keras.Model,
        X: np.ndarray,
        return_per_feature: bool = False,
        batch_size: int = int(1e10),
    ) -> np.ndarray:
        """
        Compute log probability of a batch of instances with the user defined log_prob function.
        """
        if self.sequential:
            y, X = X[:, 1:], X[:, :-1]
        else:
            y = X.copy()
        y_preds = predict_batch(X, model, batch_size=batch_size)
        logp = self.log_prob(y, y_preds).numpy()  # type: ignore
        if return_per_feature:
            return logp
        else:
            axis = tuple(np.arange(len(logp.shape))[1:])
            return np.mean(logp, axis=axis)

    def _llr(self, X: np.ndarray, return_per_feature: bool, batch_size: int = int(1e10)) -> np.ndarray:
        """
        Compute likelihood ratios.

        Parameters
        ----------
        X
            Batch of instances.
        return_per_feature
            Return likelihood ratio per feature.
        batch_size
            Batch size for the generative model evaluations.

        Returns
        -------
        Likelihood ratios.
        """
        logp_fn = self._logp if not isinstance(self.log_prob, Callable) else self._logp_alt  # type: ignore
        logp_s = logp_fn(self.dist_s, X, return_per_feature=return_per_feature, batch_size=batch_size)
        logp_b = logp_fn(self.dist_b, X, return_per_feature=return_per_feature, batch_size=batch_size)
        return logp_s - logp_b

    def score(
        self,
        X: np.ndarray,
        batch_size: int = int(1e10),
    ) -> OODScore:
        self._validate(X)
        fscore = -self._llr(X, True, batch_size=batch_size)
        iscore = -self._llr(X, False, batch_size=batch_size)
        return OODScore(iscore, fscore)



"""
Source code derived from Alibi-Detect 0.11.4
https://github.com/SeldonIO/alibi-detect/tree/v0.11.4

Original code Copyright (c) 2023 Seldon Technologies Ltd
Licensed under Apache Software License (Apache 2.0)
"""

from abc import ABC, abstractmethod
from typing import Callable, Dict, List, Literal, NamedTuple, Optional, Tuple, cast

import keras.api._v2.keras as keras
import numpy as np
import tensorflow as tf

from daml._internal.models.tensorflow.gmm import GaussianMixtureModelParams, gmm_params
from daml._internal.models.tensorflow.trainer import trainer


class OODScore(NamedTuple):
    """
    NamedTuple containing the instance and (optionally) feature score.

    Parameters
    ----------
    instance_score : np.ndarray
        Instance score of the evaluated dataset.
    feature_score : Optional[np.ndarray], default None
        Feature score, if available, of the evaluated dataset.
    """

    instance_score: np.ndarray
    feature_score: Optional[np.ndarray] = None

    def get(self, ood_type: Literal["instance", "feature"]) -> np.ndarray:
        return self.instance_score if ood_type == "instance" or self.feature_score is None else self.feature_score


class OODBase(ABC):
    def __init__(self, model: keras.Model) -> None:
        self.model = model

        self._ref_score: OODScore
        self._threshold_perc: float
        self._data_info: Optional[Tuple[tuple, type]] = None

        if not isinstance(model, keras.Model):
            raise TypeError("Model should be of type 'keras.Model'.")

    def _get_data_info(self, X: np.ndarray) -> Tuple[tuple, type]:
        if not isinstance(X, np.ndarray):
            raise TypeError("Dataset should of type: `np.ndarray`.")
        return X.shape[1:], X.dtype.type

    def _validate(self, X: np.ndarray) -> None:
        check_data_info = self._get_data_info(X)
        if self._data_info is not None and check_data_info != self._data_info:
            raise RuntimeError(f"Expect data of type: {self._data_info[1]} and shape: {self._data_info[0]}. \
                               Provided data is type: {check_data_info[1]} and shape: {check_data_info[0]}.")

    def _validate_state(self, X: np.ndarray, additional_attrs: Optional[List[str]] = None) -> None:
        attrs = ["_data_info", "_threshold_perc", "_ref_score"]
        attrs = attrs if additional_attrs is None else attrs + additional_attrs
        if not all(hasattr(self, attr) for attr in attrs) or any(getattr(self, attr) for attr in attrs) is None:
            raise RuntimeError("Metric needs to be `fit` before method call.")
        self._validate(X)

    @abstractmethod
    def score(self, X: np.ndarray, batch_size: int = int(1e10)) -> OODScore:
        """
        Compute instance and (optionally) feature level outlier scores.

        Parameters
        ----------
        X : np.ndarray
            Batch of instances.
        batch_size : int, default int(1e10)
            Batch size used when making predictions with the autoencoder.

        Returns
        -------
        Instance and feature level outlier scores.
        """

    def _threshold_score(self, ood_type: Literal["feature", "instance"] = "instance") -> np.floating:
        return np.percentile(self._ref_score.get(ood_type), self._threshold_perc)

    def fit(
        self,
        x_ref: np.ndarray,
        threshold_perc: float,
        loss_fn: Callable,
        optimizer: keras.optimizers.Optimizer,
        epochs: int,
        batch_size: int,
        verbose: bool,
    ) -> None:
        """
        Train the model and infer the threshold value.

        Parameters
        ----------
        x_ref: : np.ndarray
            Training batch.
        threshold_perc : float
            Percentage of reference data that is normal.
        loss_fn : Callable
            Loss function used for training.
        optimizer : keras.optimizers.Optimizer
            Optimizer used for training.
        epochs : int
            Number of training epochs.
        batch_size : int
            Batch size used for training.
        verbose : bool
            Whether to print training progress.
        """
        # Train the model
        trainer(
            model=self.model,
            loss_fn=loss_fn,
            x_train=x_ref,
            optimizer=optimizer,
            epochs=epochs,
            batch_size=batch_size,
            verbose=verbose,
        )

        # Infer the threshold values
        self._ref_score = self.score(x_ref, batch_size)
        self._threshold_perc = threshold_perc

    def predict(
        self,
        X: np.ndarray,
        batch_size: int = int(1e10),
        ood_type: Literal["feature", "instance"] = "instance",
    ) -> Dict[str, np.ndarray]:
        """
        Predict whether instances are out-of-distribution or not.

        Parameters
        ----------
        X
            Batch of instances.
        ood_type
            Predict out-of-distribution at the 'feature' or 'instance' level.
        batch_size
            Batch size used when making predictions with the autoencoder.

        Returns
        -------
        Dictionary containing the outlier predictions and both feature and instance level outlier scores.
        """
        self._validate_state(X)
        # compute outlier scores
        score = self.score(X, batch_size=batch_size)
        ood_pred = (score.get(ood_type) > self._threshold_score(ood_type)).astype(int)
        return {**{"is_ood": ood_pred}, **score._asdict()}


class OODGMMBase(OODBase):
    def __init__(self, model: keras.Model) -> None:
        super().__init__(model)
        self.gmm_params: GaussianMixtureModelParams

    def _validate_state(self, X: np.ndarray, additional_attrs: Optional[List[str]] = None) -> None:
        if additional_attrs is None:
            additional_attrs = ["gmm_params"]
        super()._validate_state(X, additional_attrs)

    def fit(
        self,
        x_ref: np.ndarray,
        threshold_perc: float,
        loss_fn: Callable[[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor], tf.Tensor],
        optimizer: keras.optimizers.Optimizer,
        epochs: int,
        batch_size: int,
        verbose: bool,
    ) -> None:
        # Train the model
        trainer(
            model=self.model,
            loss_fn=loss_fn,
            x_train=x_ref,
            optimizer=optimizer,
            epochs=epochs,
            batch_size=batch_size,
            verbose=verbose,
        )

        # Calculate the GMM parameters
        _, z, gamma = cast(Tuple[tf.Tensor, tf.Tensor, tf.Tensor], self.model(x_ref))
        self.gmm_params = gmm_params(z, gamma)

        # Infer the threshold values
        self._ref_score = self.score(x_ref, batch_size)
        self._threshold_perc = threshold_perc
